<?php

namespace Drupal\cmrf_form_processor;

use Drupal;
use Drupal\cmrf_form_processor\Plugin\WebformHandler\FormProcessorWebformHandler;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\DependencyInjection\DependencySerializationTrait;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Url;
use Drupal\file\Entity\File;
use Drupal\webform\Plugin\WebformElement\DateBase;
use Drupal\webform\Plugin\WebformElementAttachmentInterface;
use Drupal\webform\Utility\WebformElementHelper;
use Drupal\webform\WebformInterface;
use Drupal\webform\WebformSubmissionInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpKernel\Exception\NotFoundHttpException;
use Symfony\Component\Mime\MimeTypes;

class WebformSubmissionHandler {
  use DependencySerializationTrait;

  /**
   * The webform request handler.
   *
   * @var \Drupal\webform\WebformRequestInterface
   */
  protected $requestHandler;

  /**
   * @var \Drupal\cmrf_form_processor\Plugin\WebformHandler\FormProcessorWebformHandler
   */
  private $webformHandler;

  /** @var \Drupal\Core\Routing\RouteMatchInterface */
  private $routeMatch;

  /**
   * @var \Symfony\Component\HttpFoundation\Request
   */
  private $request;

  /**
   * @var \Drupal\cmrf_form_processor\Factory
   */
  private $factory;

  /**
   * @var array
   */
  protected $reply = [];

  /**
   * @var array
   */
  protected $defaultValues = [];

  /**
   * @var null|array
   */
  protected $calculationReply = null;

  /**
   * @var array
   */
  protected $configuration = [];

  /**
   * @var bool
   */
  protected $isSubmittedToCiviCRM = false;

  /**
   * @var \Drupal\webform\Plugin\WebformElementManager
   */
  protected $webformElementManager;

  /**
   * @var \Drupal\Core\File\FileSystem
   */
  protected $fileSystem;

  /**
   * The Webform submission exporter.
   *
   * @var \Drupal\webform\WebformSubmissionExporterInterface
   */
  protected $submissionExporter;

  public function __construct(ContainerInterface $container, Factory $factory, FormProcessorWebformHandler $webformHandler) {
    $this->factory = $factory;
    $this->webformHandler = $webformHandler;
    $this->configuration = $webformHandler->getSettings();
    $this->webformElementManager = $container->get('plugin.manager.webform.element');
    $this->submissionExporter = $container->get('webform_submission.exporter');
    $this->fileSystem = $container->get('file_system');
    $this->request = Drupal::service('request_stack')->getCurrentRequest();
    $this->requestHandler = $container->get('webform.request');
    $this->routeMatch = Drupal::routeMatch();
  }

  /**
   * Alter the form.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   The webform submission object.
   *
   * @return void
   *   This method performs alterations on the form array.
   */
  public function alterForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    if (!$form_state->isSubmitted()) {
      $values = $this->formProcessorDefaultsDefault();
      if (!empty($values) && empty($values['is_error'])) {
        unset($values['is_error']);
        foreach ($values as $field => $value) {
          $this->setDefaultValue($form, $form_state, $webform_submission, $field, $value);
        }
      }
      else {
        if (!empty($this->configuration['form_processor_require_default'])) {
          throw new NotFoundHttpException();
        }
      }
    }
    $wrapperId = $form_state->getFormObject()
        ->getFormId() . '_cmrf_form_processor';
    if (!isset($form['#prefix'])) {
      $form['#prefix'] = '';
    }
    if (!isset($form['#suffix'])) {
      $form['#suffix'] = '';
    }
    $form['#prefix'] .= '<div id="' . $wrapperId . '">';
    $form['#suffix'] = '</div>' . $form['#suffix'];
    if ($this->webformHandler->isCalculationEnabled()) {
      $defaultParams = $this->getFormProcessorDefaultParams(false);
      $element_keys = $this->factory->formProcessorCalculationParams($this->getConnection(), $this->getFormProcessor(), $defaultParams, $this->getMetaDataCacheTimeOut());
      if (count($element_keys) && $form_state->isSubmitted()) {
        $this->doCalculation($form, $form_state, $webform_submission);
      }
      $this->addDynamicDataRetrievalFields($form, $wrapperId, 'calculation', $element_keys, t('Calculating'));
    }

    // Add default handler
    $triggerDefaultElementKeys = [];
    if (isset($this->configuration['form_processor_trigger_defaults']) && is_array($this->configuration['form_processor_trigger_defaults'])) {
      $triggerDefaultElementKeys = $this->configuration['form_processor_trigger_defaults'];
    }
    if (count($triggerDefaultElementKeys)) {
      $this->addDynamicDataRetrievalFields($form, $wrapperId, 'defaults', $triggerDefaultElementKeys, t('Retrieving data'));
      if ($form_state->isSubmitted() && $form_state->isRebuilding()) {
        $triggerElement = $form_state->getTriggeringElement();
        if (isset($triggerElement['#name']) && $triggerElement['#name'] == 'cmrf_form_processor_defaults_update_button') {
          $values = $this->formProcessorDefaultsFromSubmission($form_state, $webform_submission);
          if (!empty($values) && empty($values['is_error'])) {
            unset($values['is_error']);
            $this->defaultValues = $values;
            foreach ($values as $field => $value) {
              //$this->setDefaultValue($form, $form_state, $webform_submission, $field, $value);
            }
          }
        }
      }
    }
  }

  /**
   * Validates the form submission.
   *
   * @param array $form
   *   The form array.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state object.
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   The webform submission object.
   *
   * @return void
   *   This method performs form submission validation.
   */
  public function validateForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    $triggerElement = $form_state->getTriggeringElement();
    $updateDataButtons = [
      'cmrf_form_processor_calculation_update_button',
      'cmrf_form_processor_defaults_update_button'
    ];
    if (isset($triggerElement['#name']) && in_array($triggerElement['#name'], $updateDataButtons)) {
      $form_state->clearErrors();
      $form_state->setRebuild(TRUE);
    } elseif (!empty($this->configuration['form_processor_enable_validation'])) {
      $params = $this->webformSubmissionToApiParams($webform_submission, $form_state);
      $reply = $this->factory->api(
        $this->getConnection(),
        'FormProcessorValidation',
        $this->getFormProcessor(),
        $params,
        []
      );
      foreach ($reply as $field => $msg) {
        if ($field == 'is_error') {
          continue;
        }
        $form_state->setErrorByName($field, $msg);
      }
    }
  }

  /**
   * {@inheritdoc}
   *
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   The webform submission object.
   * @param bool $update
   *   (optional) Indicates whether the webform submission is an update or not.
   * @param array $extraParams
   *   (optional) An array used to add extra parameters, for example,
   *   the results of a payment provider.
   *
   * @return void
   *   This method doesn't return a value.
   */
  public function postSave(WebformSubmissionInterface $webform_submission, $update = TRUE, $extraParams = []) {
    // $extraParams might be passed as NULL.
    $extraParams ??= [];
    $state = $webform_submission->getWebform()->getSetting('results_disabled') ? WebformSubmissionInterface::STATE_COMPLETED : $webform_submission->getState();
    if ($this->configuration['states'] && in_array($state, $this->configuration['states'])) {
      $this->sendToCivicrm($webform_submission, $extraParams);
    }
  }

  public function submitForm(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission) {
    $previousPage = null;
    $currentPage = null;
    $pages = $form_state->get('pages');
    foreach($pages as $page => $pageElement) {
      if ($form_state->get('current_page') == $page) {
        $currentPage = $page;
      } elseif (empty($currentPage)) {
        $previousPage = $page;
      }
    }
    $elementsForDefaultDataRetrievalSubmitted = [];
    // Add default handler
    $triggerDefaultElementKeys = [];
    if (isset($this->configuration['form_processor_trigger_defaults']) && is_array($this->configuration['form_processor_trigger_defaults'])) {
      $triggerDefaultElementKeys = $this->configuration['form_processor_trigger_defaults'];
    }
    if (count($triggerDefaultElementKeys)) {
      $elements = $webform_submission->getWebform()->getElementsInitializedFlattenedAndHasValue();
      foreach ($elements as $key => $element) {
        if (in_array($key, $triggerDefaultElementKeys) && isset($element['#webform_parents']) && is_array($element['#webform_parents']) && in_array($previousPage, $element['#webform_parents'])) {
          $elementsForDefaultDataRetrievalSubmitted[] = $key;
          break;
        }
      }
    }

    $canSubmitToCiviCRM = false;
    if ($currentPage == '' || $currentPage == WebformInterface::PAGE_CONFIRMATION) {
      $canSubmitToCiviCRM = !$form_state->isRebuilding();
      $state = $webform_submission->getWebform()->getSetting('results_disabled') ? WebformSubmissionInterface::STATE_COMPLETED : $webform_submission->getState();
      if (count($this->configuration['states']) && !in_array($state, $this->configuration['states'])) {
        $canSubmitToCiviCRM = false;
      }
    } elseif (count($elementsForDefaultDataRetrievalSubmitted)) {
      $values = $this->formProcessorDefaultsFromSubmission($form_state, $webform_submission);
      if (!empty($values) && empty($values['is_error'])) {
        unset($values['is_error']);
        $this->defaultValues = $values;
        foreach ($values as $field => $value) {
          $this->setDefaultValue($form, $form_state, $webform_submission, $field, $value);
        }
      }
    }

    if ($canSubmitToCiviCRM && strlen($this->configuration['form_processor_redirect_field'])) {
      $this->sendToCivicrm($webform_submission);
      if (isset($this->reply[$this->configuration['form_processor_redirect_field']])) {
        $url = $this->reply[$this->configuration['form_processor_redirect_field']];
        $form_state->setResponse(new Drupal\Core\Routing\TrustedRedirectResponse($url));
      }
    }
  }

  /**
   * {@inheritDoc}
   */
  public function postDelete(WebformSubmissionInterface $webform_submission) {
    if (in_array(WebformSubmissionInterface::STATE_DELETED, $this->configuration['states'])) {
      $this->sendToCivicrm($webform_submission);
    }
  }

  protected function doCalculation(array &$form, FormStateInterface $form_state, WebformSubmissionInterface $webform_submission, bool $force = false) {
    if (!empty($this->configuration['form_processor_enable_calculation'])) {
      $reply = $this->getCalculatedData($form_state, $webform_submission, $force);
      foreach ($reply as $field => $value) {
        $this->setDefaultValue($form, $form_state, $webform_submission, $field, $value);
      }
    }
  }

  public function getCalculatedData(FormStateInterface $form_state, WebformSubmissionInterface $webformSubmission, bool $force = false): array {
    if (empty($this->calculationReply) || $force) {
      $params = $this->webformSubmissionToApiParams($webformSubmission, $form_state);
      $this->calculationReply = $this->factory->api(
        $this->getConnection(),
        'FormProcessorCalculation',
        $this->getFormProcessor(),
        $params,
        []
      );
    }
    return $this->calculationReply;
  }

  /**
   * Get the default values for the form processor.
   *
   * @return array
   *   The default values for the form processor.
   */
  public function formProcessorDefaultsDefault(): array {
    $params = $this->getFormProcessorDefaultParams(true);
    return $this->factory->api($this->getConnection(), 'FormProcessorDefaults', $this->getFormProcessor(), $params, [], []);
  }

  /**
   * Get the default values for the form processor.
   *
   * @return array
   *   The default values for the form processor.
   */
  public function formProcessorDefaultsFromSubmission(FormStateInterface $form_state, WebformSubmissionInterface $webformSubmission): array {
    $params = $this->webformSubmissionToApiParams($webformSubmission, $form_state);
    return $this->factory->api($this->getConnection(), 'FormProcessorDefaults', $this->getFormProcessor(), $params, [], []);
  }

  /**
   * Returns the name of the connection
   *
   * @return string
   */
  public function getConnection(): string {
    return $this->configuration['connection'];
  }

  /**
   * Returns the name of the form processor
   *
   * @return string
   */
  public function getFormProcessor(): string {
    return $this->configuration['form_processor'];
  }

  /**
   * @return string
   */
  public function getMetaDataCacheTimeOut(): string {
    return $this->configuration['metadata_cache_timeout'] ?? $this->factory->getDefaultCache();
  }

  /**
   * @return array
   */
  public function getDefaultValues(): array {
    if (empty($this->defaultValues)) {
      $this->defaultValues = $this->formProcessorDefaultsDefault();
    }
    return $this->defaultValues;
  }

  /**
   * Sends the webform submission data to CiviCRM.
   *
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   The webform submission object.
   * @param array $extraParams
   *   An array used to add extra parameters, for example,
   *   the results of a payment provider.
   *
   * @return void
   *   This method doesn't return a value.
   */
  public function sendToCivicrm(WebformSubmissionInterface $webform_submission, array $extraParams = []) {
    if (!$this->isSubmittedToCiviCRM) {
      $this->isSubmittedToCiviCRM = TRUE;
      $params = $this->webformSubmissionToApiParams($webform_submission);
      if (strlen($this->configuration['form_processor_original_redirect_field'])) {
        $extraParams[$this->configuration['form_processor_original_redirect_field']] = $this->getOriginalRedirectUrl($webform_submission);
      }
      $params = array_merge($params, $extraParams);
      $this->reply = $this->factory->api(
        $this->getConnection(),
        'FormProcessor',
        $this->getFormProcessor(),
        $params,
        []
      );
    }
  }

  /**
   * @return array
   */
  public function getReply(): array {
    return $this->reply;
  }

  /**
   * Convert a webform submission to API parameters.
   *
   * @param \Drupal\webform\WebformSubmissionInterface $webform_submission
   *   The webform submission object.
   *
   * @return array
   *   The API parameters derived from the webform submission.
   */
  public function webformSubmissionToApiParams(WebformSubmissionInterface $webform_submission, FormStateInterface $form_state = null): array {
    $params = [];
    $data = $webform_submission->getData();
    if ($form_state) {
      $submittedData = $form_state->getUserInput();
      foreach($data as $key => $value) {
        if (isset($submittedData[$key])) {
          $data[$key] = $submittedData[$key];
        }
      }
      $validPages = [];
      $pages = $form_state->get('pages');
      $current_page = $form_state->get('current_page');
      foreach($pages as $page => $pageElement) {
        $validPages[] = $page;
        if ($current_page == $page) {
          break;
        }
      }
      $elements = $webform_submission->getWebform()->getElementsInitializedFlattenedAndHasValue();
      $validElements = [];
      foreach($elements as $key => $element) {
        foreach($validPages as $page) {
          if (isset($element['#webform_parents']) && is_array($element['#webform_parents']) && in_array($page, $element['#webform_parents'])) {
            $validElements[] = $key;
            break;
          }
        }
      }
      $data = array_filter($data, function($v, $k) use ($validElements) {
        return in_array($k, $validElements);
      }, ARRAY_FILTER_USE_BOTH);
    }
    $fields = $this->factory->formProcessorFields($this->getConnection(), $this->getFormProcessor());
    if (empty($this->configuration['submission_settings'])) {
      $this->configuration['submission_settings'] = ['fields' => []];
    }
    if (empty($this->configuration['submission_settings']['fields'])) {
      $this->configuration['submission_settings']['fields'] = [];
    }
    foreach ($fields as $key => $field) {
      if (array_key_exists($key, $data)) {
        if (isset($data[$key]) && in_array($this->webformHandler->getWebform()
            ->getElement($key)['#type'], [
            'webform_document_file',
            'managed_file',
            'webform_image_file',
          ])) {
          $file = File::load($data[$key]);
          $fileData = [
            'name' => $file->getFilename(),
            'mime_type' => $file->getMimeType(),
            'content' => base64_encode(file_get_contents($file->getFileUri())),
          ];
          $params[$key] = $fileData;
        }
        elseif (isset($this->configuration['submission_settings']['fields']) && isset($this->configuration['submission_settings']['fields'][$key]['format']) && $this->configuration['submission_settings']['fields'][$key]['format'] != 'raw') {
          // Use the selected format from configuration for rendering the field
          // value. We're using Webform's buildExportRecord() method for
          // rendering, as it takes care of element-specific formatting, such as
          // deciding whether to render as HTML or text.
          $element = $this->webformHandler->getWebform()->getElement($key);
          $element['#format'] = $this->configuration['submission_settings']['fields'][$key]['format'];
          $webformElement = $this->webformElementManager->getElementInstance($element);
          if ($webformElement instanceof DateBase) {
            // Make sure we export date format according to the display values.
            $params[$key] = (string) $webformElement->formatText($element, $webform_submission, ['webform' => $this->webformHandler->getWebform()]);
          }
          else {
            $export_options = $this->submissionExporter->getDefaultExportOptions();
            $export_options['webform'] = $this->webformHandler->getWebform();
            $export_record = $webformElement->buildExportRecord($element, $webform_submission, $export_options);
            $params[$key] = (string) reset($export_record);
          }
        }
        else {
          $params[$key] = $data[$key];
        }
      }
    }
    if ($this->configuration['form_processor_current_contact']) {
      $params[$this->configuration['form_processor_current_contact']] = $this->getContactId();
    }
    return $params;
  }

  /**
   * Returns an array with default params.
   *
   * @param bool $checkRequireDefault
   * @return array
   */
  public function getFormProcessorDefaultParams(bool $checkRequireDefault = true): array {
    $params = [];
    $form_processor_params = $this->configuration['form_processor_params'];
    $form_processor_params_value = $this->configuration['form_processor_params_value'] ?? [];
    if (empty($form_processor_params)) {
      if ($checkRequireDefault && !empty($this->configuration['form_processor_require_default'])) {
        throw new NotFoundHttpException();
      }
      // No form_processor params means no form defaults so return.
      return $params;
    }
    foreach ($form_processor_params as $key => $fp_param) {
      if ($fp_param == 'url' && Drupal::request()->get($key)) {
        $params[$key] = Drupal::request()->get($key);
      }
      if ($fp_param == 'current_user' && $this->getContactId()) {
        $params[$key] = $this->getContactId();
      }
      if ($fp_param == 'value' && !empty($form_processor_params_value[$key])) {
        $parsedValue = Drupal::token()->replace($form_processor_params_value[$key]);
        if (!empty($parsedValue)) {
          $params[$key] = $parsedValue;
        }
      }
    }
    if (empty($params) && !empty($checkRequireDefault && $this->configuration['form_processor_require_default'])) {
      throw new NotFoundHttpException();
    }
    return $params;
  }

  private function setDefaultValue(array &$form, FormStateInterface $formState, WebformSubmissionInterface $webformSubmission, string $field, $value) {
    $element =& WebformElementHelper::getElement($form, $field);
    if ($element) {
      $elementInstance = $this->webformElementManager->getElementInstance($element);
      if ($elementInstance instanceof WebformElementAttachmentInterface) {
        $fileDestination = 'private://cmrf_form_processor/' . Drupal::currentUser()->id() . '/' . $webformSubmission->getWebform()->id().'/'.$field . '/';
        if ($this->fileSystem->prepareDirectory($fileDestination, FileSystemInterface::CREATE_DIRECTORY | FileSystemInterface::MODIFY_PERMISSIONS)) {
          if (is_string($value) && strlen($value)) {
            $baseName = md5($this->fileSystem->basename($value));
            $fileName = $this->fileSystem->createFilename($baseName, $fileDestination);
            $uri = $this->fileSystem->saveData(file_get_contents($value), $fileName);
            $file = File::create(['uri' => $uri]);
            $file->setOwnerId(Drupal::currentUser()->id());
            $file->setTemporary();
            $mimeTypes = MimeTypes::getDefault();
            $mimeType = $mimeTypes->guessMimeType($uri);
            if ($mimeType) {
              $file->setMimeType($mimeType);
              $extensions = $mimeTypes->getExtensions($mimeType);
              if (is_array($extensions) && count($extensions)) {
                $file->setFilename($baseName.'.' . reset($extensions));
              }
            }
            $file->save();
            $element['#default_value'] = [$file->id()];
            $element['#value'] = $element['#default_value'];
            $formState->setValue($field, $element['#default_value']);
            $webformSubmission->setElementData($field, $element['#default_value']);
          }
        }
      } elseif ($element['#type'] == 'datetime') {
        try {
          $value = DrupalDateTime::createFromFormat(\DateTime::ATOM, $value);
          $element['#default_value'] = $value;
          $element['#value'] = $value;
          $formState->setValue($field, $value);
          $webformSubmission->setElementData($field, $value);
        } catch (\InvalidArgumentException $ex) {
          // is it something else than a date ignore
        }
      }
      elseif ($value !== NULL) {
        $element['#default_value'] = $value;
        $element['#value'] = $value;
        $formState->setValue($field, $value);
        $webformSubmission->setElementData($field, $value);
      }
    }
  }

  /**
   * Get the contact ID of the current user.
   *
   * @return int|null
   *   The contact ID of the current user, or null if not available.
   */
  private function getContactId() {
    $id = Drupal::currentUser()->id();
    $currentUser = Drupal::entityTypeManager()->getStorage('user')->load($id);
    return $currentUser->field_user_contact_id->value ?? NULL;
  }

  /**
   * Returns the original redirect url for this webform.
   *
   * @param WebformSubmissionInterface $webform_submission
   * @return string|null
   */
  private function getOriginalRedirectUrl(WebformSubmissionInterface $webform_submission):? string {
    $webform = $webform_submission->getWebform();
    // Get current route name, parameters, and options.
    $route_name = $this->routeMatch->getRouteName();
    $route_parameters = $this->routeMatch->getRawParameters()->all();
    $route_options = [];

    // Add token route query options.
    if (!$webform->getSetting('confirmation_exclude_token')) {
      $route_options['query']['token'] = $webform_submission->getToken();
    }

    // Add current query to route options.
    if (!$webform->getSetting('confirmation_exclude_query')) {
      $query = $this->request->query->all();
      // Remove Ajax parameters from query.
      unset($query['ajax_form'], $query['_wrapper_format']);
      if ($query) {
        $route_options['query'] = $query;
      }
    }

    $confirmation_type = $this->webformHandler->getWebformSetting('confirmation_type');
    $original_redirect_url = null;
    switch ($confirmation_type) {
      case WebformInterface::CONFIRMATION_PAGE:
        $original_redirect_url = $this->requestHandler->getUrl($webform, $webform_submission, 'webform.confirmation', $route_options);
        break;
      case WebformInterface::CONFIRMATION_URL:
      case WebformInterface::CONFIRMATION_URL_MESSAGE:
        $original_redirect_url = $this->webformHandler->getConfirmationUrl();
        if ($original_redirect_url) {
          $original_redirect_url->mergeOptions($route_options);
        }
        else {
          $route_options['query']['webform_id'] = $webform->id();
          $original_redirect_url = Url::fromRoute($route_name, $route_parameters, $route_options);
        }
        break;
    }
    if ($original_redirect_url) {
      $original_redirect_url->setAbsolute(TRUE);
      $original_redirect_url = $original_redirect_url->toString();
    }
    return $original_redirect_url;
  }

  /**
   * Add the functionality for retrieving default data when a field is changed.
   * Or for calculation.
   *
   * @param array $form
   * @param string $wrapperId
   * @param string $type
   * @param array $triggers
   * @param string $processMessage
   *
   * @return void
   */
  protected function addDynamicDataRetrievalFields(array &$form, string $wrapperId, string $type, array $triggers, string $processMessage) {
    $form['cmrf_form_processor_' . $type] = [
      '#prefix' => '<div class="js-webform-form_processor_' . $type . '" data-webform-element-keys="' . implode(',', $triggers) . '">',
      '#suffix' => '</div>',
      'update' => [
        '#type' => 'submit',
        '#value' => t('Retrieve'),
        '#ajax' => [
          'wrapper' => $wrapperId,
          'method' => 'replaceWith',
          'progress' => [
            'type' => 'throbber',
            'message' => $processMessage,
          ],
        ],
        // Disable validation, hide button, add submit button trigger class.
        '#attributes' => [
          'formnovalidate' => 'formnovalidate',
          'class' => [
            'use-ajax-submit',
            'js-hide',
            'js-webform-computed-submit',
          ],
        ],
        '#name' => 'cmrf_form_processor_' . $type . '_update_button',
      ],
      '#attached' => ['library' => ['cmrf_form_processor/webform.element.form_processor_' . $type]],
    ];
  }
}
